package SevenZip.Compression.LZMA;

import SevenZip.ICodeProgress;
import SevenZip.CoderPropID;
import SevenZip.InvalidParamException;
import SevenZip.Compression.RangeCoder.BitEncoder;
import SevenZip.Compression.LZMA.Base;

public class Encoder /* implements ISetCoderProperties, IWriteCoderProperties */ {
    public static final int EMatchFinderType_MIN  = 0;
    public static final int EMatchFinderType_BT2  = 0;
    public static final int EMatchFinderType_BT4  = 1;
    public static final int EMatchFinderType_BT4B = 2;
    public static final int EMatchFinderType_MAX  = 2;
    
    static final int kIfinityPrice = 0x7FFFFFF; // static final u_int kIfinityPrice = 0xFFFFFFF;
    
    static byte[] g_FastPos = new byte[1024];
    
    static final byte kFastSlots = 20;
    static {
        int c = 2;
        g_FastPos[0] = 0;
        g_FastPos[1] = 1;
        for (byte slotFast = 2; slotFast < kFastSlots; slotFast++) {
            int k = (1 << ((slotFast >> 1) - 1));
            for (int j = 0; j < k; j++, c++)
                g_FastPos[c] = slotFast;
        }
    }
    
    static int GetPosSlot(int pos) {
        if (pos < (1 << 10))
            return g_FastPos[pos];
        if (pos < (1 << 19))
            return (g_FastPos[pos >> 9] + 18);
        return (g_FastPos[pos >> 18] + 36);
    }
    
    static int GetPosSlot2(int pos) {
        if (pos < (1 << 16))
            return (g_FastPos[pos >> 6] + 12);
        if (pos < (1 << 25))
            return (g_FastPos[pos >> 15] + 30);
        return (g_FastPos[pos >> 24] + 48);
    }
    
    int _stateIndex = 0;
    int _previousByte;
    boolean _peviousIsMatch;
    int[] _repDistances = new int[Base.kNumRepDistances];
    
    void BaseInit() {
        _stateIndex = 0; // _state.Init();
        _previousByte = 0;
        _peviousIsMatch = false;
        for (int i = 0; i < Base.kNumRepDistances; i++)
            _repDistances[i] = 0;
    }
    
    static final int kDefaultDictionaryLogSize = 20;
    static final int kNumFastBytesDefault = 0x20;
    
    class LiteralEncoder {
        public class Encoder2 {
            BitEncoder[] m_Encoders;
            
            public void Create() {
                m_Encoders = new BitEncoder[0x300];
                for (int i = 0; i < 0x300; i++) {
                    m_Encoders[i] = new BitEncoder();
                }
            }
            
            public void Init() {
                for (int i = 0; i < 0x300; i++) {
                    m_Encoders[i].Init();
                }
            }
            
            public void Encode(SevenZip.Compression.RangeCoder.Encoder rangeEncoder, int symbol) throws java.io.IOException {
                int context = 1;
                for (int i = 7; i >= 0; i--) {
                    int bit = ((symbol >> i) & 1);
                    m_Encoders[context].Encode(rangeEncoder, bit);
                    context = (context << 1) | bit;
                }
            }
            
            public void EncodeMatched(SevenZip.Compression.RangeCoder.Encoder rangeEncoder, int matchByte, int symbol) throws java.io.IOException {
                int context = 1;
                boolean same = true;
                for (int i = 7; i >= 0; i--) {
                    int bit = ((symbol >> i) & 1);
                    int state = context;
                    if (same) {
                        int matchBit = ((matchByte >> i) & 1);
                        state += ((1 + matchBit) << 8);
                        same = (matchBit == bit);
                    }
                    m_Encoders[state].Encode(rangeEncoder, bit);
                    context = (context << 1) | bit;
                }
            }
            
            public int GetPrice(boolean matchMode, int matchByte, int symbol) {
                int price = 0;
                int context = 1;
                int i = 7;
                if (matchMode) {
                    for (; i >= 0; i--) {
                        int matchBit = (matchByte >> i) & 1;
                        int bit = (symbol >> i) & 1;
                        price += m_Encoders[((1 + matchBit) << 8) + context].GetPrice(bit);
                        context = (context << 1) | bit;
                        if (matchBit != bit) {
                            i--;
                            break;
                        }
                    }
                }
                for (; i >= 0; i--) {
                    int bit = (symbol >> i) & 1;
                    price += m_Encoders[context].GetPrice(bit);
                    context = (context << 1) | bit;
                }
                return price;
            }
        }
        
        Encoder2[] m_Coders;
        int m_NumPrevBits;
        int m_NumPosBits;
        int m_PosMask;
        
        public void Create(int numPosBits, int numPrevBits) {
            if (m_Coders != null && m_NumPrevBits == numPrevBits &&
                    m_NumPosBits == numPosBits)
                return;
            m_NumPosBits = numPosBits;
            m_PosMask = (1 << numPosBits) - 1;
            m_NumPrevBits = numPrevBits;
            int numStates = 1 << (m_NumPrevBits + m_NumPosBits);
            m_Coders = new Encoder2[numStates];
            for (int i = 0; i < numStates; i++) {
                m_Coders[i] = new Encoder2();
                m_Coders[i].Create();
            }
        }
        
        public void Init() {
            int numStates = 1 << (m_NumPrevBits + m_NumPosBits);
            for (int i = 0; i < numStates; i++)
                m_Coders[i].Init();
        }
        
        int GetState(int pos, int prevByte) {
            return ((pos & m_PosMask) << m_NumPrevBits) + ((prevByte & 0xFF) >> (8 - m_NumPrevBits)); }
        
        public Encoder2 GetSubCoder(int pos, int prevByte) {
            return m_Coders[GetState(pos, prevByte)]; }
    }
    
    class LenEncoder {
        SevenZip.Compression.RangeCoder.BitEncoder _choice = new SevenZip.Compression.RangeCoder.BitEncoder();
        SevenZip.Compression.RangeCoder.BitEncoder _choice2 = new SevenZip.Compression.RangeCoder.BitEncoder();
        SevenZip.Compression.RangeCoder.BitTreeEncoder[] _lowCoder = new SevenZip.Compression.RangeCoder.BitTreeEncoder[Base.kNumPosStatesEncodingMax]; // kNumLowBits
        SevenZip.Compression.RangeCoder.BitTreeEncoder[] _midCoder = new SevenZip.Compression.RangeCoder.BitTreeEncoder[Base.kNumPosStatesEncodingMax]; // kNumMidBits
        SevenZip.Compression.RangeCoder.BitTreeEncoder _highCoder = new SevenZip.Compression.RangeCoder.BitTreeEncoder(Base.kNumHighLenBits);
        
        public LenEncoder() {
            for (int posState = 0; posState < Base.kNumPosStatesEncodingMax; posState++) {
                _lowCoder[posState] = new SevenZip.Compression.RangeCoder.BitTreeEncoder(Base.kNumLowLenBits);
                _midCoder[posState] = new SevenZip.Compression.RangeCoder.BitTreeEncoder(Base.kNumMidLenBits);
            }
        }
        
        public void Init(int numPosStates) {
            _choice.Init();
            _choice2.Init();
            for (int posState = 0; posState < numPosStates; posState++) {
                _lowCoder[posState].Init();
                _midCoder[posState].Init();
            }
            _highCoder.Init();
        }
        
        public void Encode(SevenZip.Compression.RangeCoder.Encoder rangeEncoder, int symbol, int posState) throws java.io.IOException {
            if (symbol < Base.kNumLowLenSymbols) {
                _choice.Encode(rangeEncoder, 0);
                _lowCoder[posState].Encode(rangeEncoder, symbol);
            } else {
                symbol -= Base.kNumLowLenSymbols;
                _choice.Encode(rangeEncoder, 1);
                if (symbol < Base.kNumMidLenSymbols) {
                    _choice2.Encode(rangeEncoder, 0);
                    _midCoder[posState].Encode(rangeEncoder, symbol);
                } else {
                    _choice2.Encode(rangeEncoder, 1);
                    _highCoder.Encode(rangeEncoder, (symbol - Base.kNumMidLenSymbols));
                }
            }
        }
        
        public int GetPrice(int symbol, int posState) {
            int price = 0;
            if (symbol < Base.kNumLowLenSymbols) {
                price += _choice.GetPrice(0);
                price += _lowCoder[posState].GetPrice(symbol);
            } else {
                symbol -= Base.kNumLowLenSymbols;
                price += _choice.GetPrice(1);
                if (symbol < Base.kNumMidLenSymbols) {
                    price += _choice2.GetPrice(0);
                    price += _midCoder[posState].GetPrice(symbol);
                } else {
                    price += _choice2.GetPrice(1);
                    price += _highCoder.GetPrice((symbol - Base.kNumMidLenSymbols));
                }
            }
            return price;
        }
    };
    
    static final int kNumLenSpecSymbols = Base.kNumLowLenSymbols + Base.kNumMidLenSymbols;
    
    class LenPriceTableEncoder extends LenEncoder {
        int[] _prices = new int[Base.kNumLenSymbols << Base.kNumPosStatesBitsEncodingMax];
        int _tableSize;
        int[] _counters = new int[Base.kNumPosStatesEncodingMax];
        
        public void SetTableSize(int tableSize) { _tableSize = tableSize; }
        
        public int GetPrice(int symbol, int posState) {
            return _prices[(symbol << Base.kNumPosStatesBitsEncodingMax) + posState];
        }
        
        void UpdateTable(int posState) {
            for (int len = 0; len < _tableSize; len++)
                _prices[(len << Base.kNumPosStatesBitsEncodingMax) + posState] = super.GetPrice(len, posState);
            _counters[posState] = _tableSize;
        }
        
        public void UpdateTables(int numPosStates) {
            for (int posState = 0; posState < numPosStates; posState++)
                UpdateTable(posState);
        }
        
        public void Encode(SevenZip.Compression.RangeCoder.Encoder rangeEncoder, int symbol, int posState) throws java.io.IOException {
            super.Encode(rangeEncoder, symbol, posState);
            if (--_counters[posState] == 0)
                UpdateTable(posState);
        }
    }
    
    static final int kNumOpts = 1 << 12;
    class Optimal {
        int stateIndex = 0; // public State State;
        
        public boolean Prev1IsChar;
        public boolean Prev2;
        
        public int PosPrev2;
        public int BackPrev2;
        
        public int Price;
        public int PosPrev;         // posNext;
        public int BackPrev;
        
        public int Backs0;
        public int Backs1;
        public int Backs2;
        public int Backs3;
        
        public void MakeAsChar() { BackPrev = -1 /* 0xFFFFFFFF */ ; Prev1IsChar = false; }
        public void MakeAsShortRep() { BackPrev = 0; ; Prev1IsChar = false; }
        public boolean IsShortRep() { return (BackPrev == 0); }
    };
    Optimal[] _optimum = new Optimal[kNumOpts];
    
    SevenZip.Compression.LZ.BinTree _matchFinder = null; // SevenZip.Compression.LZ.IMatchFinder _matchFinder
    SevenZip.Compression.RangeCoder.Encoder _rangeEncoder = new SevenZip.Compression.RangeCoder.Encoder();
    
    SevenZip.Compression.RangeCoder.BitEncoder[] _isMatch = new SevenZip.Compression.RangeCoder.BitEncoder[Base.kNumStates << Base.kNumPosStatesBitsMax];
    SevenZip.Compression.RangeCoder.BitEncoder[] _isRep = new SevenZip.Compression.RangeCoder.BitEncoder[Base.kNumStates];
    SevenZip.Compression.RangeCoder.BitEncoder[] _isRepG0 = new SevenZip.Compression.RangeCoder.BitEncoder[Base.kNumStates];
    SevenZip.Compression.RangeCoder.BitEncoder[] _isRepG1 = new SevenZip.Compression.RangeCoder.BitEncoder[Base.kNumStates];
    SevenZip.Compression.RangeCoder.BitEncoder[] _isRepG2 = new SevenZip.Compression.RangeCoder.BitEncoder[Base.kNumStates];
    SevenZip.Compression.RangeCoder.BitEncoder[] _isRep0Long = new SevenZip.Compression.RangeCoder.BitEncoder[Base.kNumStates << Base.kNumPosStatesBitsMax];
    
    SevenZip.Compression.RangeCoder.BitTreeEncoder[] _posSlotEncoder = new SevenZip.Compression.RangeCoder.BitTreeEncoder[Base.kNumLenToPosStates]; // kNumPosSlotBits
    SevenZip.Compression.RangeCoder.BitEncoder[] _posEncoders = new SevenZip.Compression.RangeCoder.BitEncoder[Base.kNumFullDistances - Base.kEndPosModelIndex];
    SevenZip.Compression.RangeCoder.BitTreeEncoder _posAlignEncoder = new SevenZip.Compression.RangeCoder.BitTreeEncoder(Base.kNumAlignBits);
    
    LenPriceTableEncoder _lenEncoder = new LenPriceTableEncoder();
    LenPriceTableEncoder _repMatchLenEncoder = new LenPriceTableEncoder();
    
    LiteralEncoder _literalEncoder = new LiteralEncoder();
    
    int[] _matchDistances = new int[Base.kMatchMaxLen + 1];
    boolean _fastMode = false;
    boolean _maxMode = false;
    int _numFastBytes = kNumFastBytesDefault;
    int _longestMatchLength;
    int _additionalOffset;
    
    int _optimumEndIndex;
    int _optimumCurrentIndex;
    
    boolean _longestMatchWasFound;
    
    int[] _posSlotPrices = new int[Base.kDistTableSizeMax << Base.kNumLenToPosStatesBits];
    int[] _distancesPrices = new int[Base.kNumFullDistances << Base.kNumLenToPosStatesBits];
    int[] _alignPrices = new int[Base.kAlignTableSize];
    int _alignPriceCount;
    
    int _distTableSize = (kDefaultDictionaryLogSize * 2);
    
    int _posStateBits = 2;
    int _posStateMask = (4 - 1);
    int _numLiteralPosStateBits = 0;
    int _numLiteralContextBits = 3;
    
    int _dictionarySize = (1 << kDefaultDictionaryLogSize);
    int _dictionarySizePrev = -1; // 0xFFFFFFFF;
    int _numFastBytesPrev = -1; // 0xFFFFFFFF;
    
    long lastPosSlotFillingPos;
    long nowPos64;
    boolean _finished;
    
    java.io.InputStream _inStream;
    int _matchFinderType = EMatchFinderType_BT4;
    boolean _writeEndMark;
    boolean _needReleaseMFStream;
    
    void Create() {
        if (_matchFinder == null) {
            SevenZip.Compression.LZ.BinTree bt = new SevenZip.Compression.LZ.BinTree();
            int numHashBytes = 4;
            boolean big = false;
            switch (_matchFinderType) {
                case EMatchFinderType_BT2:
                    numHashBytes = 2;
                    break;
                case EMatchFinderType_BT4:
                    break;
                case EMatchFinderType_BT4B:
                    big = true;
                    break;
            }
            bt.SetType(numHashBytes, big);
            _matchFinder = bt;
        }
        _literalEncoder.Create(_numLiteralPosStateBits, _numLiteralContextBits);
        
        if (_dictionarySize == _dictionarySizePrev && _numFastBytesPrev == _numFastBytes)
            return;
        _matchFinder.Create(_dictionarySize, kNumOpts, _numFastBytes,
                (Base.kMatchMaxLen - _numFastBytes));
        _dictionarySizePrev = _dictionarySize;
        _numFastBytesPrev = _numFastBytes;
    }
    
    public Encoder() {
        for (int i = 0; i < kNumOpts; i++)
            _optimum[i] = new Optimal();
        for (int i = 0; i < Base.kNumLenToPosStates; i++)
            _posSlotEncoder[i] = new SevenZip.Compression.RangeCoder.BitTreeEncoder(Base.kNumPosSlotBits);
        for (int i = 0; i < Base.kNumStates; i++) {
            for (int j = 0; j <= _posStateMask; j++) {
                int complexState = (i << Base.kNumPosStatesBitsMax) + j;
                _isMatch[complexState] = new SevenZip.Compression.RangeCoder.BitEncoder();
                _isRep0Long[complexState] = new SevenZip.Compression.RangeCoder.BitEncoder();
            }
            _isRep[i] = new SevenZip.Compression.RangeCoder.BitEncoder();
            _isRepG0[i] = new SevenZip.Compression.RangeCoder.BitEncoder();
            _isRepG1[i] = new SevenZip.Compression.RangeCoder.BitEncoder();
            _isRepG2[i] = new SevenZip.Compression.RangeCoder.BitEncoder();
        }
        for (int i = 0; i < Base.kNumFullDistances - Base.kEndPosModelIndex; i++) {
            _posEncoders[i] = new SevenZip.Compression.RangeCoder.BitEncoder();
        }
    }
    
    void SetWriteEndMarkerMode(boolean writeEndMarker) {
        _writeEndMark = writeEndMarker;
    }
    void Init() {
        BaseInit();
        _rangeEncoder.Init();
        
        int i;
        for (i = 0; i < Base.kNumStates; i++) {
            for (int j = 0; j <= _posStateMask; j++) {
                int complexState = (i << Base.kNumPosStatesBitsMax) + j;
                _isMatch[complexState].Init();
                _isRep0Long[complexState].Init();
            }
            _isRep[i].Init();
            _isRepG0[i].Init();
            _isRepG1[i].Init();
            _isRepG2[i].Init();
        }
        _literalEncoder.Init();
        for (i = 0; i < Base.kNumLenToPosStates; i++) {
            _posSlotEncoder[i].Init();
        }
        for (i = 0; i < Base.kNumFullDistances - Base.kEndPosModelIndex; i++) {
            _posEncoders[i].Init();
        }
        
        _lenEncoder.Init(1 << _posStateBits);
        _repMatchLenEncoder.Init(1 << _posStateBits);
        
        _posAlignEncoder.Init();
        
        _longestMatchWasFound = false;
        _optimumEndIndex = 0;
        _optimumCurrentIndex = 0;
        _additionalOffset = 0;
    }
    
    int ReadMatchDistances() throws java.io.IOException {
        int lenRes = _matchFinder.GetLongestMatch(_matchDistances);
        if (lenRes == _numFastBytes)
            lenRes += _matchFinder.GetMatchLen(lenRes, _matchDistances[lenRes],Base.kMatchMaxLen - lenRes);
        _additionalOffset++;
        _matchFinder.MovePos();
        return lenRes;
    }
    
    void MovePos(int num) throws java.io.IOException {
        for (; num > 0; num--) {
            _matchFinder.DummyLongestMatch();
            _matchFinder.MovePos();
            _additionalOffset++;
        }
    }
    
    int GetRepLen1Price(int state_Index, int posState) {
        return _isRepG0[state_Index].GetPrice(0) +
                _isRep0Long[(state_Index << Base.kNumPosStatesBitsMax) + posState].GetPrice(0);
    }
    
    int GetRepPrice(int repIndex, int len, int state_Index, int posState) {
        int price = _repMatchLenEncoder.GetPrice(len - Base.kMatchMinLen, posState);
        if (repIndex == 0) {
            price += _isRepG0[state_Index].GetPrice(0);
            price += _isRep0Long[(state_Index << Base.kNumPosStatesBitsMax) + posState].GetPrice(1);
        } else {
            price += _isRepG0[state_Index].GetPrice(1);
            if (repIndex == 1)
                price += _isRepG1[state_Index].GetPrice(0);
            else {
                price += _isRepG1[state_Index].GetPrice(1);
                price += _isRepG2[state_Index].GetPrice(repIndex - 2);
            }
        }
        return price;
    }
    
    int GetPosLenPrice(int pos, int len, int posState) {
        if (len == 2 && pos >= 0x80)
            return kIfinityPrice;
        int price;
        int lenToPosState = Base.GetLenToPosState(len);
        if (pos < Base.kNumFullDistances)
            price = _distancesPrices[(pos << Base.kNumLenToPosStatesBits) + lenToPosState];
        else
            price = _posSlotPrices[(GetPosSlot2(pos) << Base.kNumLenToPosStatesBits) + lenToPosState] +
                    _alignPrices[pos & Base.kAlignMask];
        return price + _lenEncoder.GetPrice(len - Base.kMatchMinLen, posState);
    }
    
    int Backward(int [] backRes, int cur) {
        _optimumEndIndex = cur;
        int posMem = _optimum[cur].PosPrev;
        int backMem = _optimum[cur].BackPrev;
        do
        {
            if (_optimum[cur].Prev1IsChar) {
                _optimum[posMem].MakeAsChar();
                _optimum[posMem].PosPrev = posMem - 1;
                if (_optimum[cur].Prev2) {
                    _optimum[posMem - 1].Prev1IsChar = false;
                    _optimum[posMem - 1].PosPrev = _optimum[cur].PosPrev2;
                    _optimum[posMem - 1].BackPrev = _optimum[cur].BackPrev2;
                }
            }
            int posPrev = posMem;
            int backCur = backMem;
            
            backMem = _optimum[posPrev].BackPrev;
            posMem = _optimum[posPrev].PosPrev;
            
            _optimum[posPrev].BackPrev = backCur;
            _optimum[posPrev].PosPrev = cur;
            cur = posPrev;
        }
        while (cur > 0);
        backRes[0] = _optimum[0].BackPrev;
        _optimumCurrentIndex = _optimum[0].PosPrev;
        return _optimumCurrentIndex;
    }
    
    
    int GetOptimum(int position, int [] backRes) throws java.io.IOException {
        if (_optimumEndIndex != _optimumCurrentIndex) {
            int lenRes = _optimum[_optimumCurrentIndex].PosPrev - _optimumCurrentIndex;
            backRes[0] = _optimum[_optimumCurrentIndex].BackPrev;
            _optimumCurrentIndex = _optimum[_optimumCurrentIndex].PosPrev;
            return lenRes;
        }
        _optimumCurrentIndex = 0;
        _optimumEndIndex = 0; // test it;
        
        int lenMain;
        if (!_longestMatchWasFound) {
            lenMain = ReadMatchDistances();
        } else {
            lenMain = _longestMatchLength;
            _longestMatchWasFound = false;
        }
        
        int[] reps = new int[Base.kNumRepDistances];
        int[] repLens = new int[Base.kNumRepDistances];
        int repMaxIndex = 0;
        int i;
        for (i = 0; i < Base.kNumRepDistances; i++) {
            reps[i] = _repDistances[i];
            repLens[i] = _matchFinder.GetMatchLen(0 - 1, reps[i], Base.kMatchMaxLen);
            if (i == 0 || repLens[i] > repLens[repMaxIndex])
                repMaxIndex = i;
        }
        if (repLens[repMaxIndex] > _numFastBytes) {
            backRes[0] = repMaxIndex;
            int lenRes = repLens[repMaxIndex];
            MovePos(lenRes - 1);
            return lenRes;
        }
        
        if (lenMain > _numFastBytes) {
            int backMain = (lenMain < _numFastBytes) ? _matchDistances[lenMain] :
                _matchDistances[_numFastBytes];
            backRes[0] = backMain + Base.kNumRepDistances;
            MovePos(lenMain - 1);
            return lenMain;
        }
        int currentByte = _matchFinder.GetIndexByte(0 - 1);
        
        _optimum[0].stateIndex = _stateIndex;
        
        int matchByte;
        
        matchByte = _matchFinder.GetIndexByte(0 - _repDistances[0] - 1 - 1);
        
        int posState = (position & _posStateMask);
        
        _optimum[1].Price = _isMatch[(_stateIndex << Base.kNumPosStatesBitsMax) + posState].GetPrice(0) +
                _literalEncoder.GetSubCoder(position, _previousByte).GetPrice(_peviousIsMatch, matchByte, currentByte);
        _optimum[1].MakeAsChar();
        
        _optimum[1].PosPrev = 0;
        
        _optimum[0].Backs0 = reps[0];
        _optimum[0].Backs1 = reps[1];
        _optimum[0].Backs2 = reps[2];
        _optimum[0].Backs3 = reps[3];
        
        
        int matchPrice = _isMatch[(_stateIndex << Base.kNumPosStatesBitsMax) + posState].GetPrice(1);
        int repMatchPrice = matchPrice + _isRep[_stateIndex].GetPrice(1);
        
        if (matchByte == currentByte) {
            int shortRepPrice = repMatchPrice + GetRepLen1Price(_stateIndex, posState);
            if (shortRepPrice < _optimum[1].Price) {
                _optimum[1].Price = shortRepPrice;
                _optimum[1].MakeAsShortRep();
            }
        }
        if (lenMain < 2) {
            backRes[0] = _optimum[1].BackPrev;
            return 1;
        }
        
        
        int normalMatchPrice = matchPrice +
                _isRep[_stateIndex].GetPrice(0);
        
        if (lenMain <= repLens[repMaxIndex])
            lenMain = 0;
        
        int len;
        for (len = 2; len <= lenMain; len++) {
            _optimum[len].PosPrev = 0;
            _optimum[len].BackPrev = _matchDistances[len] + Base.kNumRepDistances;
            _optimum[len].Price = normalMatchPrice +
                    GetPosLenPrice(_matchDistances[len], len, posState);
            _optimum[len].Prev1IsChar = false;
        }
        
        if (lenMain < repLens[repMaxIndex])
            lenMain = repLens[repMaxIndex];
        
        for (; len <= lenMain; len++)
            _optimum[len].Price = kIfinityPrice;
        
        for (i = 0; i < Base.kNumRepDistances; i++) {
            int repLen = repLens[i];
            for (int lenTest = 2; lenTest <= repLen; lenTest++) {
                int curAndLenPrice = repMatchPrice + GetRepPrice(i, lenTest, _stateIndex, posState);
                Optimal optimum = _optimum[lenTest];
                if (curAndLenPrice < optimum.Price) {
                    optimum.Price = curAndLenPrice;
                    optimum.PosPrev = 0;
                    optimum.BackPrev = i;
                    optimum.Prev1IsChar = false;
                }
            }
        }
        
        int cur = 0;
        int lenEnd = lenMain;
        
        while (true) {
            cur++;
            if (cur == lenEnd) {
                return Backward(backRes, cur);
            }
            position++;
            int posPrev = _optimum[cur].PosPrev;
            int stateIndex; // State state;
            if (_optimum[cur].Prev1IsChar) {
                posPrev--;
                if (_optimum[cur].Prev2) {
                    stateIndex = _optimum[_optimum[cur].PosPrev2].stateIndex;
                    if (_optimum[cur].BackPrev2 < Base.kNumRepDistances)
                        stateIndex = Base.UpdateRep(stateIndex);
                    else
                        stateIndex = Base.UpdateMatch(stateIndex);
                } else {
                    stateIndex = _optimum[posPrev].stateIndex;
                }
                stateIndex = Base.UpdateChar(stateIndex);
            } else {
                stateIndex = _optimum[posPrev].stateIndex;
            }
            boolean prevWasMatch;
            if (posPrev == cur - 1) {
                if (_optimum[cur].IsShortRep()) {
                    prevWasMatch = true;
                    stateIndex = Base.UpdateShortRep(stateIndex);
                } else {
                    prevWasMatch = false;
                    stateIndex = Base.UpdateChar(stateIndex);
                }
            } else {
                prevWasMatch = true;
                int pos;
                if (_optimum[cur].Prev1IsChar && _optimum[cur].Prev2) {
                    posPrev = _optimum[cur].PosPrev2;
                    pos = _optimum[cur].BackPrev2;
                    stateIndex = Base.UpdateRep(stateIndex);
                } else {
                    pos = _optimum[cur].BackPrev;
                    if (pos < Base.kNumRepDistances)
                        stateIndex = Base.UpdateRep(stateIndex);
                    else
                        stateIndex = Base.UpdateMatch(stateIndex);
                }
                Optimal opt = _optimum[posPrev];
                if (pos < Base.kNumRepDistances) {
                    if (pos == 0) {
                        reps[0] = opt.Backs0;
                        reps[1] = opt.Backs1;
                        reps[2] = opt.Backs2;
                        reps[3] = opt.Backs3;
                    } else if (pos == 1) {
                        reps[0] = opt.Backs1;
                        reps[1] = opt.Backs0;
                        reps[2] = opt.Backs2;
                        reps[3] = opt.Backs3;
                    } else if (pos == 2) {
                        reps[0] = opt.Backs2;
                        reps[1] = opt.Backs0;
                        reps[2] = opt.Backs1;
                        reps[3] = opt.Backs3;
                    } else {
                        reps[0] = opt.Backs3;
                        reps[1] = opt.Backs0;
                        reps[2] = opt.Backs1;
                        reps[3] = opt.Backs2;
                    }
                } else {
                    reps[0] = (pos - Base.kNumRepDistances);
                    reps[1] = opt.Backs0;
                    reps[2] = opt.Backs1;
                    reps[3] = opt.Backs2;
                }
            }
            _optimum[cur].stateIndex = stateIndex;
            _optimum[cur].Backs0 = reps[0];
            _optimum[cur].Backs1 = reps[1];
            _optimum[cur].Backs2 = reps[2];
            _optimum[cur].Backs3 = reps[3];
            int newLen;
            newLen = ReadMatchDistances();
            if (newLen > _numFastBytes) {
                _longestMatchLength = newLen;
                _longestMatchWasFound = true;
                return Backward(backRes, cur);
            }
            int curPrice = _optimum[cur].Price;
            
            currentByte = _matchFinder.GetIndexByte(0 - 1);
            matchByte = _matchFinder.GetIndexByte(0 - reps[0] - 1 - 1);
            
            posState = (position & _posStateMask);
            
            int curAnd1Price = curPrice +
                    _isMatch[(stateIndex << Base.kNumPosStatesBitsMax) + posState].GetPrice(0) +
                    _literalEncoder.GetSubCoder(position, _matchFinder.GetIndexByte(0 - 2)).
                    GetPrice(prevWasMatch, matchByte, currentByte);
            
            Optimal nextOptimum = _optimum[cur + 1];
            
            boolean nextIsChar = false;
            if (curAnd1Price < nextOptimum.Price) {
                nextOptimum.Price = curAnd1Price;
                nextOptimum.PosPrev = cur;
                nextOptimum.MakeAsChar();
                nextIsChar = true;
            }
            
            matchPrice = curPrice + _isMatch[(stateIndex << Base.kNumPosStatesBitsMax) + posState].GetPrice(1);
            repMatchPrice = matchPrice + _isRep[stateIndex].GetPrice(1);
            
            if (matchByte == currentByte &&
                    !(nextOptimum.PosPrev < cur && nextOptimum.BackPrev == 0)) {
                int shortRepPrice = repMatchPrice + GetRepLen1Price(stateIndex, posState);
                if (shortRepPrice <= nextOptimum.Price) {
                    nextOptimum.Price = shortRepPrice;
                    nextOptimum.PosPrev = cur;
                    nextOptimum.MakeAsShortRep();
                }
            }
            
            int numAvailableBytes = _matchFinder.GetNumAvailableBytes() + 1;
            numAvailableBytes = Math.min(kNumOpts - 1 - cur, numAvailableBytes);
            
            if (numAvailableBytes < 2)
                continue;
            if (numAvailableBytes > _numFastBytes)
                numAvailableBytes = _numFastBytes;
            if (numAvailableBytes >= 3 && !nextIsChar) {
                int backOffset = reps[0] + 1;
                int temp;
                for (temp = 1; temp < numAvailableBytes; temp++)
                    // if (data[temp] != data[(size_t)temp - backOffset])
                    if (_matchFinder.GetIndexByte(temp - 1) != _matchFinder.GetIndexByte(temp - backOffset - 1))
                        break;
                int lenTest2 = temp - 1;
                if (lenTest2 >= 2) {
                    int stateIndex2 = stateIndex;
                    stateIndex2 = Base.UpdateChar(stateIndex2);
                    int posStateNext = (position + 1) & _posStateMask;
                    int nextRepMatchPrice = curAnd1Price +
                            _isMatch[(stateIndex2 << Base.kNumPosStatesBitsMax) + posStateNext].GetPrice(1) +
                            _isRep[stateIndex2].GetPrice(1);
                    // for (; lenTest2 >= 2; lenTest2--)
                    {
                        while (lenEnd < cur + 1 + lenTest2)
                            _optimum[++lenEnd].Price = kIfinityPrice;
                        int curAndLenPrice = nextRepMatchPrice + GetRepPrice(
                                0, lenTest2, stateIndex2, posStateNext);
                        Optimal optimum = _optimum[cur + 1 + lenTest2];
                        if (curAndLenPrice < optimum.Price) {
                            optimum.Price = curAndLenPrice;
                            optimum.PosPrev = cur + 1;
                            optimum.BackPrev = 0;
                            optimum.Prev1IsChar = true;
                            optimum.Prev2 = false;
                        }
                    }
                }
            }
            for (int repIndex = 0; repIndex < Base.kNumRepDistances; repIndex++) {
                int backOffset = reps[repIndex] + 1;
                int lenTest;
                for (lenTest = 0; lenTest < numAvailableBytes; lenTest++)
                    // if (data[lenTest] != data[(size_t)lenTest - backOffset])
                    if (_matchFinder.GetIndexByte(lenTest - 1) != _matchFinder.GetIndexByte(lenTest - backOffset - 1))
                        break;
                for (; lenTest >= 2; lenTest--) {
                    while (lenEnd < cur + lenTest)
                        _optimum[++lenEnd].Price = kIfinityPrice;
                    int curAndLenPrice = repMatchPrice + GetRepPrice(repIndex, lenTest, stateIndex, posState);
                    Optimal optimum = _optimum[cur + lenTest];
                    if (curAndLenPrice < optimum.Price) {
                        optimum.Price = curAndLenPrice;
                        optimum.PosPrev = cur;
                        optimum.BackPrev = repIndex;
                        optimum.Prev1IsChar = false;
                    }
                }
            }
            
            if (newLen > numAvailableBytes)
                newLen = numAvailableBytes;
            if (newLen >= 2) {
                if (newLen == 2 && _matchDistances[2] >= 0x80)
                    continue;
                normalMatchPrice = matchPrice +
                        _isRep[stateIndex].GetPrice(0);
                while (lenEnd < cur + newLen)
                    _optimum[++lenEnd].Price = kIfinityPrice;
                
                for (int lenTest = newLen; lenTest >= 2; lenTest--) {
                    int curBack = _matchDistances[lenTest];
                    int curAndLenPrice = normalMatchPrice + GetPosLenPrice(curBack, lenTest, posState);
                    Optimal optimum = _optimum[cur + lenTest];
                    if (curAndLenPrice < optimum.Price) {
                        optimum.Price = curAndLenPrice;
                        optimum.PosPrev = cur;
                        optimum.BackPrev = curBack + Base.kNumRepDistances;
                        optimum.Prev1IsChar = false;
                    }
                    
                    if (_maxMode) {
                        int backOffset = curBack + 1;
                        int temp;
                        for (temp = lenTest + 1; temp < numAvailableBytes; temp++)
                            // if (data[temp] != data[(size_t)temp - backOffset])
                            if (_matchFinder.GetIndexByte(temp - 1) != _matchFinder.GetIndexByte(temp - backOffset - 1))
                                break;
                        int lenTest2 = temp - (lenTest + 1);
                        if (lenTest2 >= 2) {
                            int stateIndex2 = stateIndex;
                            stateIndex2 = Base.UpdateMatch(stateIndex2);
                            int posStateNext = (position + lenTest) & _posStateMask;
                            int curAndLenCharPrice = curAndLenPrice +
                                    _isMatch[(stateIndex2 << Base.kNumPosStatesBitsMax) + posStateNext].GetPrice(0) +
                                    _literalEncoder.GetSubCoder(position + lenTest,
                                    // data[(size_t)lenTest - 1],
                                    _matchFinder.GetIndexByte(lenTest - 1 - 1)).
                                    GetPrice(true,
                                    // data[(size_t)lenTest - backOffset], data[lenTest]
                                    _matchFinder.GetIndexByte(lenTest - backOffset - 1),
                                    _matchFinder.GetIndexByte(lenTest - 1)
                                    );
                            stateIndex2 = Base.UpdateChar(stateIndex2);
                            posStateNext = (position + lenTest + 1) & _posStateMask;
                            int nextMatchPrice = curAndLenCharPrice + _isMatch[(stateIndex2 << Base.kNumPosStatesBitsMax) + posStateNext].GetPrice(1);
                            int nextRepMatchPrice = nextMatchPrice + _isRep[stateIndex2].GetPrice(1);
                            
                            // for(; lenTest2 >= 2; lenTest2--)
                            {
                                int offset = lenTest + 1 + lenTest2;
                                while (lenEnd < cur + offset)
                                    _optimum[++lenEnd].Price = kIfinityPrice;
                                curAndLenPrice = nextRepMatchPrice + GetRepPrice(
                                        0, lenTest2, stateIndex2, posStateNext);
                                optimum = _optimum[cur + offset];
                                if (curAndLenPrice < optimum.Price) {
                                    optimum.Price = curAndLenPrice;
                                    optimum.PosPrev = cur + lenTest + 1;
                                    optimum.BackPrev = 0;
                                    optimum.Prev1IsChar = true;
                                    optimum.Prev2 = true;
                                    optimum.PosPrev2 = cur;
                                    optimum.BackPrev2 = curBack + Base.kNumRepDistances;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    static final int kDif = 7;
    boolean ChangePair(int smallDist, int bigDist) {
        return (smallDist < ((1) << (32 - kDif)) && bigDist >= (smallDist << kDif));
    }
    
    int GetOptimumFast(int position, int [] backRes) throws java.io.IOException {
        int lenMain;
        if (!_longestMatchWasFound) {
            lenMain = ReadMatchDistances();
        } else {
            lenMain = _longestMatchLength;
            _longestMatchWasFound = false;
        }
        int[] repLens = new int[Base.kNumRepDistances];
        int repMaxIndex = 0;
        for (int i = 0; i < Base.kNumRepDistances; i++) {
            repLens[i] = _matchFinder.GetMatchLen(0 - 1, _repDistances[i], Base.kMatchMaxLen);
            if (i == 0 || repLens[i] > repLens[repMaxIndex])
                repMaxIndex = i;
        }
        if (repLens[repMaxIndex] >= _numFastBytes) {
            backRes[0] = repMaxIndex;
            int lenRes = repLens[repMaxIndex];
            MovePos(lenRes - 1);
            return lenRes;
        }
        if (lenMain >= _numFastBytes) {
            backRes[0] = _matchDistances[_numFastBytes] + Base.kNumRepDistances;
            MovePos(lenMain - 1);
            return lenMain;
        }
        while (lenMain > 2) {
            if (!ChangePair(_matchDistances[lenMain - 1], _matchDistances[lenMain]))
                break;
            lenMain--;
        }
        if (lenMain == 2 && _matchDistances[2] >= 0x80)
            lenMain = 1;
        
        int backMain = _matchDistances[lenMain];
        if (repLens[repMaxIndex] >= 2) {
            if (repLens[repMaxIndex] + 1 >= lenMain ||
                    repLens[repMaxIndex] + 2 >= lenMain && (backMain > (1 << 12))) {
                backRes[0] = repMaxIndex;
                int lenRes = repLens[repMaxIndex];
                MovePos(lenRes - 1);
                return lenRes;
            }
        }
        
        if (lenMain >= 2) {
            _longestMatchLength = ReadMatchDistances();
            if (_longestMatchLength >= 2 &&
                    (
                    (_longestMatchLength >= lenMain &&
                    _matchDistances[lenMain] < backMain) ||
                    _longestMatchLength == lenMain + 1 &&
                    !ChangePair(backMain, _matchDistances[_longestMatchLength]) ||
                    _longestMatchLength > lenMain + 1 ||
                    _longestMatchLength + 1 >= lenMain &&
                    ChangePair(_matchDistances[lenMain - 1], backMain)
                    )
                    ) {
                _longestMatchWasFound = true;
                backRes[0] = -1; // 0xFFFFFFFF;
                return 1;
            }
            for (int i = 0; i < Base.kNumRepDistances; i++) {
                int repLen = _matchFinder.GetMatchLen(0 - 1, _repDistances[i], Base.kMatchMaxLen);
                if (repLen >= 2 && repLen + 1 >= lenMain) {
                    _longestMatchWasFound = true;
                    backRes[0] = -1; // 0xFFFFFFFF;
                    return 1;
                }
            }
            backRes[0] = backMain + Base.kNumRepDistances;
            MovePos(lenMain - 2);
            return lenMain;
        }
        backRes[0] = -1; // 0xFFFFFFFF;
        return 1;
    }
    
    void WriteEndMarker(int posState) throws java.io.IOException {
        if (!_writeEndMark)
            return;
        
        _isMatch[(_stateIndex << Base.kNumPosStatesBitsMax) + posState].Encode(_rangeEncoder, 1);
        _isRep[_stateIndex].Encode(_rangeEncoder, 0);
        _stateIndex = Base.UpdateMatch(_stateIndex);
        int len = Base.kMatchMinLen; // kMatchMaxLen;
        _lenEncoder.Encode(_rangeEncoder, len - Base.kMatchMinLen, posState);
        int posSlot = (1 << Base.kNumPosSlotBits) - 1;
        int lenToPosState = Base.GetLenToPosState(len);
        _posSlotEncoder[lenToPosState].Encode(_rangeEncoder, posSlot);
        int footerBits = 30;
        int posReduced = (1 << footerBits) - 1;
        _rangeEncoder.EncodeDirectBits((posReduced >> Base.kNumAlignBits), footerBits - Base.kNumAlignBits);
        _posAlignEncoder.ReverseEncode(_rangeEncoder, posReduced & Base.kAlignMask);
    }
    
    void Flush(int nowPos) throws java.io.IOException {
        ReleaseMFStream();
        WriteEndMarker(nowPos & _posStateMask);
        _rangeEncoder.FlushData();
        _rangeEncoder.FlushStream();
    }
    
    public void CodeOneBlock(long [] inSize, long [] outSize, boolean [] finished) throws java.io.IOException {
        inSize[0] = 0;
        outSize[0] = 0;
        finished[0] = true;
        
        if (_inStream != null) {
            _matchFinder.Init(_inStream);
            _needReleaseMFStream = true;
            _inStream = null;
        }
        
        if (_finished)
            return;
        _finished = true;
        
        long progressPosValuePrev = nowPos64;
        if (nowPos64 == 0) {
            if (_matchFinder.GetNumAvailableBytes() == 0) {
                Flush((int)nowPos64);
                return;
            }
            int len; // it's not used
            len = ReadMatchDistances();
            int posState = (int)(nowPos64) & _posStateMask;
            _isMatch[(_stateIndex << Base.kNumPosStatesBitsMax) + posState].Encode(_rangeEncoder, 0);
            _stateIndex = Base.UpdateChar(_stateIndex);
            int curByte = _matchFinder.GetIndexByte(0 - _additionalOffset);
            _literalEncoder.GetSubCoder((int)(nowPos64), _previousByte).Encode(_rangeEncoder, curByte);
            _previousByte = curByte;
            _additionalOffset--;
            nowPos64++;
        }
        if (_matchFinder.GetNumAvailableBytes() == 0) {
            Flush((int)nowPos64);
            return;
        }
        int [] pos = new int[1];
        while (true) {
            int posState = ((int)nowPos64) & _posStateMask;
            int len;
            if (_fastMode) {
                len = GetOptimumFast((int)nowPos64, pos);
            } else {
                len = GetOptimum((int)nowPos64, pos);
            }
            
            int complexState = (_stateIndex << Base.kNumPosStatesBitsMax) + posState;
            if (len == 1 && pos[0] == -1 /* 00xFFFFFFFF */ ) {
                _isMatch[complexState].Encode(_rangeEncoder, 0);
                _stateIndex = Base.UpdateChar(_stateIndex);
                int curByte = _matchFinder.GetIndexByte(0 - _additionalOffset);
                LiteralEncoder.Encoder2 subCoder = _literalEncoder.GetSubCoder((int)nowPos64, _previousByte);
                if (_peviousIsMatch) {
                    int matchByte = _matchFinder.GetIndexByte(0 - _repDistances[0] - 1 - _additionalOffset);
                    subCoder.EncodeMatched(_rangeEncoder, matchByte, curByte);
                } else {
                    subCoder.Encode(_rangeEncoder, curByte);
                }
                _previousByte = curByte;
                _peviousIsMatch = false;
            } else {
                _peviousIsMatch = true;
                _isMatch[complexState].Encode(_rangeEncoder, 1);
                if (pos[0] < Base.kNumRepDistances) {
                    _isRep[_stateIndex].Encode(_rangeEncoder, 1);
                    if (pos[0] == 0) {
                        _isRepG0[_stateIndex].Encode(_rangeEncoder, 0);
                        if (len == 1) {
                            _isRep0Long[complexState].Encode(_rangeEncoder, 0);
                        } else {
                            _isRep0Long[complexState].Encode(_rangeEncoder, 1);
                        }
                    } else {
                        _isRepG0[_stateIndex].Encode(_rangeEncoder, 1);
                        if (pos[0] == 1) {
                            _isRepG1[_stateIndex].Encode(_rangeEncoder, 0);
                        } else {
                            _isRepG1[_stateIndex].Encode(_rangeEncoder, 1);
                            _isRepG2[_stateIndex].Encode(_rangeEncoder, pos[0] - 2);
                        }
                    }
                    if (len == 1) {
                        _stateIndex = Base.UpdateShortRep(_stateIndex);
                    } else {
                        _repMatchLenEncoder.Encode(_rangeEncoder, len - Base.kMatchMinLen, posState);
                        _stateIndex = Base.UpdateRep(_stateIndex);
                    }
                    
                    
                    int distance = _repDistances[pos[0]];
                    if (pos[0] != 0) {
                        for (int i = pos[0]; i >= 1; i--)
                            _repDistances[i] = _repDistances[i - 1];
                        _repDistances[0] = distance;
                    }
                } else {
                    _isRep[_stateIndex].Encode(_rangeEncoder, 0);
                    _stateIndex = Base.UpdateMatch(_stateIndex);
                    _lenEncoder.Encode(_rangeEncoder, len - Base.kMatchMinLen, posState);
                    pos[0] -= Base.kNumRepDistances;
                    int posSlot = GetPosSlot(pos[0]);
                    int lenToPosState = Base.GetLenToPosState(len);
                    _posSlotEncoder[lenToPosState].Encode(_rangeEncoder, posSlot);
                    
                    if (posSlot >= Base.kStartPosModelIndex) {
                        int footerBits = ((posSlot >> 1) - 1);
                        int baseVal = ((2 | (posSlot & 1)) << footerBits);
                        int posReduced = pos[0] - baseVal;
                        
                        if (posSlot < Base.kEndPosModelIndex) {
                            SevenZip.Compression.RangeCoder.BitTreeEncoder.ReverseEncode(_posEncoders,
                                    baseVal - posSlot - 1, _rangeEncoder, footerBits, posReduced);
                        } else {
                            _rangeEncoder.EncodeDirectBits((posReduced >> Base.kNumAlignBits), footerBits - Base.kNumAlignBits);
                            _posAlignEncoder.ReverseEncode(_rangeEncoder, posReduced & Base.kAlignMask);
                            if (!_fastMode)
                                if (--_alignPriceCount == 0)
                                    FillAlignPrices();
                        }
                    }
                    int distance = pos[0];
                    for (int i = Base.kNumRepDistances - 1; i >= 1; i--)
                        _repDistances[i] = _repDistances[i - 1];
                    _repDistances[0] = distance;
                }
                _previousByte = _matchFinder.GetIndexByte(len - 1 - _additionalOffset);
            }
            _additionalOffset -= len;
            nowPos64 += len;
            if (!_fastMode) {
                if (nowPos64 - lastPosSlotFillingPos >= (1 << 9)) {
                    FillPosSlotPrices();
                    FillDistancesPrices();
                    lastPosSlotFillingPos = nowPos64;
                }
            }
            if (_additionalOffset == 0) {
                inSize[0] = nowPos64;
                outSize[0] = _rangeEncoder.GetProcessedSizeAdd();
                if (_matchFinder.GetNumAvailableBytes() == 0) {
                    Flush((int)nowPos64);
                    return;
                }
                
                if (nowPos64 - progressPosValuePrev >= (1 << 12)) {
                    _finished = false;
                    finished[0] = false;
                    return;
                }
            }
        }
    }
    
    void ReleaseMFStream() {
        if (_matchFinder != null && _needReleaseMFStream) {
            _matchFinder.ReleaseStream();
            _needReleaseMFStream = false;
        }
    }
    
    void SetOutStream(java.io.OutputStream outStream) { _rangeEncoder.SetStream(outStream); }
    void ReleaseOutStream() { _rangeEncoder.ReleaseStream(); }
    
    void ReleaseStreams() {
        ReleaseMFStream();
        ReleaseOutStream();
    }
    
    void SetStreams(java.io.InputStream inStream, java.io.OutputStream outStream) {
        _inStream = inStream;
        _finished = false;
        Create();
        SetOutStream(outStream);
        Init();
        
        // CCoderReleaser releaser(this);
        
                        /*
                        if (_matchFinder->GetNumAvailableBytes() == 0)
                                return Flush();
                         */
        
        if (!_fastMode) {
            FillPosSlotPrices();
            FillDistancesPrices();
            FillAlignPrices();
        }
        
        _lenEncoder.SetTableSize(_numFastBytes);
        _lenEncoder.UpdateTables(1 << _posStateBits);
        _repMatchLenEncoder.SetTableSize(_numFastBytes);
        _repMatchLenEncoder.UpdateTables(1 << _posStateBits);
        
        lastPosSlotFillingPos = 0;
        nowPos64 = 0;
    }
    
    public void Code(java.io.InputStream inStream, java.io.OutputStream outStream, ICodeProgress progress) throws java.io.IOException {
        _needReleaseMFStream = false;
        try {
            SetStreams(inStream, outStream);
            long [] processedInSize = new long[1];
            long [] processedOutSize = new long[1];
            boolean [] finished = new boolean[1];
            while (true) {
                CodeOneBlock(processedInSize, processedOutSize, finished);
                if (finished[0])
                    return;
                if (progress != null) {
                    progress.SetProgress(processedInSize[0], processedOutSize[0]);
                }
            }
        } finally {
            ReleaseStreams();
        }
    }
    
    static final int kDicLogSizeMaxCompress = 28;
    public void SetCoderProperties(int[] propIDs, int[] properties) throws InvalidParamException {
        for (int i = 0; i < properties.length; i++) {
            int prop = properties[i];
            switch (propIDs[i]) {
                case CoderPropID.NumFastBytes:
                {
                    int numFastBytes = prop;
                    if (numFastBytes < 2 || numFastBytes > Base.kMatchMaxLen)
                        throw new InvalidParamException();
                    _numFastBytes = numFastBytes;
                    break;
                }
                case CoderPropID.Algorithm:
                {
                    int maximize = prop;
                    _fastMode = (maximize == 0);
                    _maxMode = (maximize >= 2);
                    break;
                }
                case CoderPropID.MatchFinder:
                {
                    int matchFinderIndexPrev = _matchFinderType;
                    int m = prop;
                    if ((m < EMatchFinderType_MIN) ||(m > EMatchFinderType_MAX))
                        throw new InvalidParamException();
                    _matchFinderType = m;
                    if (_matchFinder != null && matchFinderIndexPrev != _matchFinderType) {
                        _dictionarySizePrev = -1; // 0xFFFFFFFF;
                        _matchFinder = null;
                    }
                    break;
                }
                case CoderPropID.DictionarySize:
                {
                    int dictionarySize = prop;
                    if (dictionarySize < (1 << Base.kDicLogSizeMin) ||
                            dictionarySize > (1 << kDicLogSizeMaxCompress))
                        throw new InvalidParamException();
                    _dictionarySize = dictionarySize;
                    int dicLogSize;
                    for (dicLogSize = 0; dicLogSize < kDicLogSizeMaxCompress; dicLogSize++)
                        if (dictionarySize <= (1 << dicLogSize))
                            break;
                    _distTableSize = dicLogSize * 2;
                    break;
                }
                case CoderPropID.PosStateBits:
                {
                    int v = prop;
                    if (v < 0 || v > Base.kNumPosStatesBitsEncodingMax)
                        throw new InvalidParamException();
                    _posStateBits = v;
                    _posStateMask = (1 << _posStateBits) - 1;
                    break;
                }
                case CoderPropID.LitPosBits:
                {
                    int v = prop;
                    if (v < 0 || v > Base.kNumLitPosStatesBitsEncodingMax)
                        throw new InvalidParamException();
                    _numLiteralPosStateBits = v;
                    break;
                }
                case CoderPropID.LitContextBits:
                {
                    int v = prop;
                    if ((v < 0) || (v > Base.kNumLitContextBitsMax))
                        throw new InvalidParamException();
                    _numLiteralContextBits = v;
                    break;
                }
                case CoderPropID.EndMarker:
                {
                    if (prop != 0) SetWriteEndMarkerMode(true);
                    else SetWriteEndMarkerMode(false);
                    break;
                }
                default:
                    throw new InvalidParamException();
            }
        }
    }
    
    static final int kPropSize = 5;
    public void WriteCoderProperties(java.io.OutputStream outStream) throws java.io.IOException {
        byte[] properties = new byte[kPropSize];
        properties[0] = (byte)((_posStateBits * 5 + _numLiteralPosStateBits) * 9 + _numLiteralContextBits);
        for (int i = 0; i < 4; i++)
            properties[1 + i] = (byte)(_dictionarySize >> (8 * i));
        outStream.write(properties, 0, kPropSize);
    }
    
    void FillPosSlotPrices() {
        for (int lenToPosState = 0; lenToPosState < Base.kNumLenToPosStates; lenToPosState++) {
            int posSlot;
            for (posSlot = 0; posSlot < Base.kEndPosModelIndex && posSlot < _distTableSize; posSlot++)
                _posSlotPrices[(posSlot << Base.kNumLenToPosStatesBits) + lenToPosState] = _posSlotEncoder[lenToPosState].GetPrice(posSlot);
            for (; posSlot < _distTableSize; posSlot++)
                _posSlotPrices[(posSlot << Base.kNumLenToPosStatesBits) + lenToPosState] = _posSlotEncoder[lenToPosState].GetPrice(posSlot) +
                        ((((posSlot >> 1) - 1) - Base.kNumAlignBits) << SevenZip.Compression.RangeCoder.BitEncoder.kNumBitPriceShiftBits);
        }
    }
    
    void FillDistancesPrices() {
        for (int lenToPosState = 0; lenToPosState < Base.kNumLenToPosStates; lenToPosState++) {
            int i;
            for (i = 0; i < Base.kStartPosModelIndex; i++)
                _distancesPrices[(i << Base.kNumLenToPosStatesBits) + lenToPosState] = _posSlotPrices[(i << Base.kNumLenToPosStatesBits) + lenToPosState];
            for (; i < Base.kNumFullDistances; i++) {
                int posSlot = GetPosSlot(i);
                int footerBits = (posSlot >> 1) - 1;
                int baseVal = ((2 | (posSlot & 1)) << footerBits);
                
                _distancesPrices[(i << Base.kNumLenToPosStatesBits) + lenToPosState] = _posSlotPrices[(posSlot << Base.kNumLenToPosStatesBits) + lenToPosState] +
                        SevenZip.Compression.RangeCoder.BitTreeEncoder.ReverseGetPrice(_posEncoders,
                        baseVal - posSlot - 1, footerBits, i - baseVal);
            }
        }
    }
    
    void FillAlignPrices() {
        for (int i = 0; i < Base.kAlignTableSize; i++)
            _alignPrices[i] = _posAlignEncoder.ReverseGetPrice(i);
        _alignPriceCount = Base.kAlignTableSize;
    }
}
