package SevenZip;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStream;

public class LzmaAlone {
    
    static void PrintHelp() {
        System.out.println("\nUsage:  LZMA <e|d|b> [<switches>...] inputFile outputFile\n" +
                "  e: encode file\n" +
                "  d: decode file\n" +
                "  b: Benchmark\n" +
                "<Switches>\n" +
                "  -a{N}:  set compression mode - [0, 2], default: 2 (max)\n" +
                "  -d{N}:  set dictionary - [0,28], default: 23 (8MB)\n" +
                "  -fb{N}: set number of fast bytes - [5, 255], default: 128\n" +
                "  -lc{N}: set number of literal context bits - [0, 8], default: 3\n" +
                "  -lp{N}: set number of literal pos bits - [0, 4], default: 0\n" +
                "  -pb{N}: set number of pos bits - [0, 4], default: 2\n" +
                "  -mf{MF_ID}: set Match Finder: [bt2, bt4, bt4b, default: bt4\n" +
                "  -eos:   write End Of Stream marker\n"
                );
    }
    
    static int IncorrectCommand() throws Exception {
        throw (new Exception("Command line error"));
    }
    
    public static int main2(String args[])  throws Exception {
        if (args.length < 1) {
            PrintHelp();
            return 0;
        }
        
        String command = args[0].toLowerCase();
        
        boolean dictionaryIsDefined = false;
        int dictionary = 1 << 21;
        int posStateBits = 2;
        int litContextBits = 3; // for normal files
        // int litContextBits = 0; // for 32-bit data
        int litPosBits = 0;
        // int litPosBits = 2; // for 32-bit data
        int algorithm = 2;
        int numFastBytes = 128;
        String smf ="bt4";
        
        int eos = 0; // false;
        
        int paramIndex = 1;
        
        for (int i = 1; i < args.length ; i++) {
            String flags = args[i].toUpperCase();
            if (flags.charAt(0) == '-') {
                if (flags.compareToIgnoreCase("--") == 0) {
                    paramIndex = i + 1;
                    break;
                } else if (flags.equals("-H")) {
                    PrintHelp();
                    return 0;
                } else if (flags.equals("-?")) {
                    PrintHelp();
                    return 0;
                } else if (flags.startsWith("-A")) { // -a{N}:  set compression mode - [0, 2], default: 2 (max)
                    try {
                        algorithm = Integer.parseInt(flags.substring(2));
                    } catch (NumberFormatException e) {
                        IncorrectCommand();
                    }
                } else if (flags.startsWith("-D")) { //  -d{N}:  set dictionary - [0,28], default: 23 (8MB)
                    try {
                        dictionary = 1 << Integer.parseInt(flags.substring(2));
                        dictionaryIsDefined = true;
                    } catch (NumberFormatException e) {
                        IncorrectCommand();
                    }
                } else if (flags.startsWith("-FB")) { //  -fb{N}: set number of fast bytes - [5, 255], default: 128
                    try {
                        numFastBytes = Integer.parseInt(flags.substring(3));
                    } catch (NumberFormatException e) {
                        IncorrectCommand();
                    }
                } else if (flags.startsWith("-LC")) { //  -lc{N}: set number of literal context bits - [0, 8], default: 3
                    try {
                        litContextBits = Integer.parseInt(flags.substring(3));
                    } catch (NumberFormatException e) {
                        IncorrectCommand();
                    }
                } else if (flags.startsWith("-LP")) { //  -lp{N}: set number of literal pos bits - [0, 4], default: 0
                    try {
                        litPosBits = Integer.parseInt(flags.substring(3));
                    } catch (NumberFormatException e) {
                        IncorrectCommand();
                    }
                } else if (flags.startsWith("-PB")) { //  -pb{N}: set number of pos bits - [0, 4], default: 2
                    try {
                        posStateBits = Integer.parseInt(flags.substring(3));
                    } catch (NumberFormatException e) {
                        IncorrectCommand();
                    }
                } else if (flags.startsWith("-MF")) { //  -mf{MF_ID}: set Match Finder: [bt2, bt4, bt4b, default: bt4
                    smf = flags.substring(3);
                    smf = smf.toLowerCase();
                } else if (flags.equals("-eos")) { // write End Of Stream marker : true
                    eos = 1;
                } else {
                    IncorrectCommand();
                }
            } else {
                paramIndex = i;
                break;
            }
        }
        
        if (command.equals("b")) {
            
            final int kNumDefaultItereations = 10;
            int numIterations = kNumDefaultItereations;
            if (paramIndex < args.length)
                try {
                    numIterations = Integer.parseInt(args[paramIndex]);
                } catch (NumberFormatException e) {
                    IncorrectCommand();
                }
            return LzmaBench.LzmaBenchmark(numIterations, dictionary, smf.equals("bt4"));
        }
        
        boolean encodeMode = false;
        if (command.equals("e"))
            encodeMode = true;
        else if (command.equals("d"))
            encodeMode = false;
        else
            IncorrectCommand();
        
        if (paramIndex + 2 != args.length)
            IncorrectCommand();
        
        File fileInput = new File(args[paramIndex]);
        FileInputStream inStream = new FileInputStream(fileInput);
        OutputStream outStream = new FileOutputStream(new File(args[paramIndex+1]));
        
        if (encodeMode) {
            if (!dictionaryIsDefined)
                dictionary = 1 << 23;
            
            int [] propIDs =
            {
                CoderPropID.DictionarySize,
                        CoderPropID.PosStateBits,
                        CoderPropID.LitContextBits,
                        CoderPropID.LitPosBits,
                        CoderPropID.Algorithm,
                        CoderPropID.NumFastBytes,
                        CoderPropID.MatchFinder,
                        CoderPropID.EndMarker
            };
            
            int mf=-1;
            if (smf.equals("bt2")) mf = SevenZip.Compression.LZMA.Encoder.EMatchFinderType_BT2;
            else if (smf.equals("bt4")) mf = SevenZip.Compression.LZMA.Encoder.EMatchFinderType_BT4;
            else if (smf.equals("bt4b")) mf = SevenZip.Compression.LZMA.Encoder.EMatchFinderType_BT4B;
            else IncorrectCommand();
            
            int[] properties =
            {
                dictionary,
                        posStateBits,
                        litContextBits,
                        litPosBits,
                        algorithm,
                        numFastBytes,
                        mf,
                        eos
            };
            
            SevenZip.Compression.LZMA.Encoder encoder = new SevenZip.Compression.LZMA.Encoder();
            encoder.SetCoderProperties(propIDs, properties);
            encoder.WriteCoderProperties(outStream);
            long fileSize;
            if (eos == 1) // (eos || stdInMode)
                fileSize = -1;
            else
                fileSize = fileInput.length(); // inStream.Length;
            for (int i = 0; i < 8; i++)
                outStream.write((byte)(fileSize >> (8 * i)));
            
            encoder.Code(inStream, outStream, null);
            inStream.close();
            outStream.close();
        } else if (command.equals("d")) {
            byte[] properties = new byte[5];
            if (inStream.read(properties, 0, 5) != 5)
                throw (new Exception("input .lzma is too short"));
            SevenZip.Compression.LZMA.Decoder decoder = new SevenZip.Compression.LZMA.Decoder();
            decoder.SetDecoderProperties(properties);
            long outSize = 0;
            for (int i = 0; i < 8; i++) {
                int v = inStream.read();
                if (v < 0)
                    throw (new Exception("Can't Read 1"));
                outSize |= ((long)v) << (8 * i);
            }
            java.io.BufferedInputStream  inStream0  = new java.io.BufferedInputStream(inStream);
            java.io.BufferedOutputStream outStream0 = new java.io.BufferedOutputStream(outStream);
            decoder.Code(inStream0, outStream0, outSize, null);
            inStream0.close();
            outStream0.close();
        } else
            throw (new Exception("Command Error"));
        
        return 0;
    }
    
    public static void main(String[] args) {
        try {
            long temps = System.currentTimeMillis();
            int code = main2(args);
            temps = System.currentTimeMillis() - temps;
            System.out.println("time : " + temps + " ms");
            System.exit(code);
        } catch (Exception e) {
            System.out.println(e.getMessage());
            System.exit(1);
        }
    }
}

